from math import sqrt,inf
import numpy as np
import matplotlib.pyplot as plt
import random
import seaborn as sns
import pandas as pd

def euclidian_distance(v1,v2):
    """v1 and v2 are vectors of the same dimension"""
    result = 0
    for i in range(len(v1)):
        result += (v1[i] - v2[i])**2
    return sqrt(result)

def dot_product(v1, v2):
    product = 0
    for i in range(len(v1)):
        product += v1[i]*v2[i]
    return product

def euclidian_norm(v):
    return sqrt(dot_product(v,v))

def cosine_distance(v1, v2):
    norm_v1 = euclidian_norm(v1)
    norm_v2 = euclidian_norm(v2)
    return 1 - dot_product(v1,v2)/(norm_v1 * norm_v2)

def covariance_matrix(C):
    C = np.array(C)
    C_centered = C - np.mean(C, axis=0)
    covariance_matrix = np.dot(C_centered.T, C_centered) / (C.shape[0] - 1)
    return covariance_matrix

def distance_mahalanobis(v1, v2, cov_matrix):
    covariance_inv = np.linalg.inv(cov_matrix)
    delta = np.array(v1) - np.array(v2)
    dist = np.sqrt(np.dot(np.dot(delta.T, covariance_inv), delta))
    return dist

def class_center(C):
    "center of the class C"
    avg = [0]*len(C[0])
    for i in range(len(C[0])):
        sum = 0
        for j in range(len(C)):
            sum += C[j][i]
        sum = sum/len(C)
        avg[i] = sum
    return avg

def intra_class_distance(C, dist, cov_matrix=[], mahalanobis=False):
    """C is an array of arrays of floats"""
    centre_c = class_center(C)
    maximum = 0
    if not mahalanobis:
        for elt in C:
            distance = dist(centre_c, elt)
            if distance > maximum:
                maximum = distance
    else:
        for elt in C:
            distance = dist(centre_c, elt, cov_matrix)
            if distance > maximum:
                maximum = distance
    return maximum

def distance_orientee(C1,C2,dist, cov_matrix=[], mahalanobis=False):
    center2 = class_center(C2)
    minimum = inf
    if not mahalanobis:
        for elt in C1:
            distance = dist(elt,center2)
            if distance < minimum:
                minimum = distance
    else:
        for elt in C1:
            distance = dist(elt,center2, cov_matrix)
            if distance < minimum:
                minimum = distance
    if minimum < 0.00001:
        return inf
    return minimum

def inter_class_distance(C1, C2, dist, cov_matrix=[], mahalanobis=False):
    return min(distance_orientee(C1,C2,dist,cov_matrix, mahalanobis), distance_orientee(C2,C1,dist,cov_matrix, mahalanobis))

def overlap(C1, C2, dist, cov_matrix=[], mahalanobis=False):
    intra_c1 = intra_class_distance(C1, dist, cov_matrix, mahalanobis)
    intra_c2 = intra_class_distance(C2, dist, cov_matrix, mahalanobis)
    inter = inter_class_distance(C1, C2, dist, cov_matrix, mahalanobis)
    return (intra_c1 + intra_c2)/(2*inter)

def isOverlapped(C1,C2,dist):
    return overlap(C1,C2,dist) >= 1

def turnDataIntoFloat(string_array):
    new_array = [0]*len(string_array)
    for i in range(len(string_array)):
        new_array[i] = float(string_array[i])
    return new_array

def getData(set_of_genes):
    """creates a dictionnary containing all the data in arrays for each tumor"""
    classes = {"PRAD":[], "LUAD":[], "COAD":[], "KIRC":[], "BRCA":[]}
    labels = open("labels.csv", 'r')
    data = open("data.csv", 'r')
    labels.readline()
    data.readline()
    tumor_per_sample = labels.readlines()
    data_per_sample = data.readlines()
    for i, sample in enumerate(data_per_sample):
        sample_data = sample.split(',')
        sample_data = turnDataIntoFloat([sample_data[k] for k in set_of_genes])
        tumor_name = tumor_per_sample[i].split(',')[1].strip('\n')
        classes[tumor_name].append(sample_data)
    return classes

bijection = {0:"PRAD", 1:"LUAD", 2:"COAD", 3:"KIRC", 4:"BRCA"}

def calculateOverlapArray(dist, classes, mahalanobis=False):
    """creates an array that summarizes all the overlapses between classes of tumors"""
    nb_tumor = len(classes)
    overlap_array = [[0,0,0,0,0], [0,0,0,0,0], [0,0,0,0,0], [0,0,0,0,0], [0,0,0,0,0]]
    for i in range(nb_tumor):
        for j in range(i,nb_tumor):
            #calculating overlap between i and j
            cov_matrix = covariance_matrix(classes[bijection[j]])
            for k in range(len(cov_matrix)):
                cov_matrix[k][k] += 0.001
            overlap_array[i][j] = overlap(classes[bijection[i]], classes[bijection[j]], dist, cov_matrix, mahalanobis)
    return overlap_array

def mean_overlaps(dist, classes, mahalanobis = False):
    overlap_array = calculateOverlapArray(dist, classes, mahalanobis)
    mean = 0
    n = len(overlap_array)
    k = 0
    for i in range(0, n-1):
        for j in range(i+1, n):
            mean += overlap_array[i][j]
            k += 1
    mean /= k
    return mean

def plot_nuage_de_points(**classes):
    size = len(classes)
    assert(size<=5) #il y a au plus 5 classes.
    
    colors = ['red', 'blue', 'green', 'yellow', 'purple']
    for i, key in enumerate(classes):
        # Séparation des composantes pour chaque type
        x, y = [v[0] for v in classes[key]], [v[1] for v in classes[key]]
        plt.scatter(x, y, color=colors[i], label=key, s=20)

    # Ajout de labels et de légende
    plt.xlabel('Gène 2106')
    plt.ylabel('Gène 319')
    plt.title('Nuage de points avec les gènes 2106 et 319')
    plt.legend()

    plt.show()
    return 0

def plot_histogramme(**classes):
    size = len(classes)
    assert(size<=5) #il y a au plus 5 classes.
    
    colors = ['red', 'blue', 'green', 'yellow', 'purple']
    dataframes = []
    for i, key in enumerate(classes):
        df_classe = pd.DataFrame(classes[key], columns=['Gène 1', 'Gène 2'])
        df_classe['Classe'] = key
        dataframes.append(df_classe)
    df = pd.concat(dataframes)

    # Tracer un jointplot en utilisant seaborn
    g = sns.jointplot(data=df, x="Gène 1", y="Gène 2", hue="Classe", kind="scatter", palette="Set1")

    # Ajouter des labels et un titre
    plt.suptitle("Distribution des gènes avec jointplot", y=1.02)
    plt.show()
    return 0

def minimize_average_overlaps(dist):
    """find the 2 best genes that minimize the average overlaps using the euclidian distance"""
    last_gene = 20531
    min_mean_overlaps = inf
    gene1 = -1
    gene2 = -1
    nb_test = 10000
    for k in range(nb_test):
        i = random.randint(1,last_gene)
        j = i
        while j == i:
            j = random.randint(1,last_gene)
        print("testing ", i, "and ", j, " : test ", k)
        data = getData(i,j)
        mean_test = mean_overlaps(euclidian_distance, data)
        if mean_test < min_mean_overlaps and np.mean([t[0] for t in data["BRCA"]])>=0.2 and np.mean([t[1] for t in data["BRCA"]])>=0.2:   #checking they are not genes that are almost always at 0
            min_mean_overlaps = mean_test
            gene1 = i
            gene2 = j
        print("current min = ", min_mean_overlaps, " with gene1 = ", gene1, " and gene2 = ", gene2)
    classes = getData(gene1,gene2)
    plot_nuage_de_points(PRAD=classes["PRAD"],LUAD=classes["LUAD"],COAD=classes["COAD"],KIRC=classes["KIRC"], BRCA=classes["BRCA"])
    return gene1, gene2, min_mean_overlaps


#classes = getData([i for i in range(1,20532)])
#plot_histogramme(PRAD=classes["PRAD"],LUAD=classes["LUAD"],COAD=classes["COAD"],KIRC=classes["KIRC"], BRCA=classes["BRCA"])
#plot_nuage_de_points(PRAD=classes["PRAD"],LUAD=classes["LUAD"],COAD=classes["COAD"],KIRC=classes["KIRC"], BRCA=classes["BRCA"])

#print(calculateOverlapArray(distance_mahalanobis, classes, True))
#print(calculateOverlapArray(euclidian_distance, classes))
#print(calculateOverlapArray(cosine_distance, classes))


